{
@abstract(basic doc generator object)
@author(Ivan Montes Velencoso (senbei@teleline.es))
@author(Marco Schmidt (marcoschmidt@geocities.com))
@author(Philippe Jean Dit Bailleul (jdb@abacom.com))
@author(Rodrigo Urubatan Ferreira Jardim (rodrigo@netscape.net))
@created(30 Aug 1998)
@lastmod(20 Apr 2000)

GenDoc contains the basic documentation generator object @link(TDocGenerator).
It is not sufficient by itself but the basis for all generators that produce
documentation in a specific format like HTML or LaTex.
They override @link(TDocGenerator)'s virtual methods.
}
unit gendoc;

interface

uses
  arrays, chars, filestre, items, numbers, objects, parsing, streams, texts;

const
  { number of overview files that pasdoc generates for
    multiple-document-formats like @link(HTML) }
  NUM_OVERVIEW_FILES = 7;
  { names of all overview files, extensions not included }
  OverviewFileNames: Array[0..NUM_OVERVIEW_FILES - 1] of String[8] =
    ('allunits', 'allclass', 'alltypes', 'allvaria', 'allconst', 'allfuncs',
     'allident');
  { integer constants for all texts that are needed in pasdoc's output }
  TEXT_AUTHOR = 1;
  TEXT_AUTHORS = 2;
  TEXT_CIO_SUMMARY = 3;
  TEXT_CLASS = 4;
  TEXT_CLASSES = 5;
  TEXT_CIO = 6;
  TEXT_CONSTANTS = 7;
  TEXT_CREATED = 8;
  TEXT_DESCRIPTION = 9;
  TEXT_FIELDS = 10;
  TEXT_FUNCTIONS_AND_PROCEDURES = 11;
  TEXT_HEADLINE_UNITS = 12;
  TEXT_HEADLINE_CIO = 13;
  TEXT_HEADLINE_TYPES = 14;
  TEXT_HEADLINE_VARIABLES = 15;
  TEXT_HEADLINE_CONSTANTS = 16;
  TEXT_HEADLINE_FUNCSPROCS = 17;
  TEXT_HEADLINE_IDENTIFIERS = 18;
  TEXT_INTERFACE = 19;
  TEXT_LASTMOD = 20;
  TEXT_METHODS = 21;
  TEXT_NAME = 22;
  TEXT_NONE = 23;
  TEXT_OBJECT = 24;
  TEXT_OBJECTS = 25;
  TEXT_OVERVIEW = 26;
  TEXT_OVERVIEW_UNITS = 27;
  TEXT_OVERVIEW_CIO = 28;
  TEXT_OVERVIEW_TYPES = 29;
  TEXT_OVERVIEW_VARIABLES = 30;
  TEXT_OVERVIEW_CONSTANTS = 31;
  TEXT_OVERVIEW_FUNCSPROCS = 32;
  TEXT_OVERVIEW_IDENTIFIERS = 33;
  TEXT_PROPERTIES = 34;
  TEXT_SUMMARY_OF_PROPERTIES = 35;
  TEXT_TYPES = 36;
  TEXT_UNIT = 37;
  TEXT_VARIABLES = 38;
  TEXT_WARNING_AUTOMATIC = 39;
  TEXT_SUMMARY_OF_METHODS = 40;
  TEXT_DECLARATION = 41;
  TEXT_TYPE = 42;

type
  { pointer to @link(TDocGenerator) }
  PDocGenerator = ^TDocGenerator;
  { @abstract(basic documentation generator object)
    @author(Marco Schmidt (marcoschmidt@geocities.com))
    This abstract object will do the complete process of writing
	documentation files.
    It will be given the collection of units that was the result of the
    parsing process and a configuration object that was created from default
    values and program parameters.
    Depending on the output format, one or more files may be created (HTML
    will create several, Tex only one). }
  TDocGenerator = object(TObject)
    { destination directory for documentation; must include terminating
      forward slash or backslash so that valid file names can be created
      by concatenating DestDir and a pathless file name }
    DestDir: String;
    KeepLineFeeds: Boolean;
    { the (human) output language of the documentation file(s);
      one of the LANG_xxx constants, e.g. @link(LANG_ENGLISH);
      default language is @link(DEFAULT_LANGUAGE) }
    Language: TLanguage;
    { if true, no link to pasdoc homepage will be included at the bottom of
      HTML files;
      default is false }
    NoGeneratorInfo: Boolean;
    { the output stream that is currently written to; depending on the
      output format, more than one output stream will be necessary to
      store all documentation }
    Stream: POutputStream;
    { list of all units that were successfully parsed }
    Units: PItemCollection;
    constructor Init;
    { Creates anchors and links for all items in all units. }
    procedure BuildLinks;
    { Checks if D is assigned and empty - if so, disposes of D and sets it to
      nil.
      If there are characters in D, it is checked whether at least one
      non-whitespace character is present - if all characters are whitespace,
      disposes of D. }
    procedure CheckForEmptyDescription(var D: PText);
    { If field @link(Stream) is assigned, it is disposed and set to nil. }
    procedure CloseStream;
    { Character conversion function - heirs are supposed to override this
      version to return escaped versions for special characters, like a
      backslash in @link(Tex) or an ampersand character in @link(HTML).
      This version returns an empty string for all input characters. }
    function ConvertChar(C: Char): String; virtual;
    { Calls @link(ConvertChar) for each character in S, thus assembling a
      string that is returned and can be written to the documentation file. }
    function ConvertString(S: String): String;
    { Abstract function to be overwritten in descendants. 
      This function is supposed to return a reference to an item, that is the
      name combined with some linking information like a hyperlink element in
      HTML or a page number in Tex. }
    function CreateLink(Item: PItem): String; virtual;
    { If field @link(Stream) still exists (<> nil), it is closed.
      Then, a new output stream in the destination directory with given
      name and file extension typical for this document format is created and
      assigned to Stream.
      No path or extension should therefore be in Name.
      Typical values for Name would be 'Objects' or 'AllUnits'.
      Returns true if creation was successful, false otherwise. }
    function CreateStream(Name: String): Boolean;
    { Must be overwritten.
      From an item name and its link, this creates a language-specific
      reference to that item. }
    function CreateReferencedLink(ItemName, Link: String): String; virtual;
    { Takes description D, expands links (using Item), converts
      output-specific characters.
      Returns true on success, false otherwise (not enough memory?). }
    function ExpandDescription(Item: PItem; var D: PText): Boolean;
    { Calls @link(ExpandDescription) for each item in each unit of
      @link(Units). }
    procedure ExpandDescriptions;
    { Searches for an email address in string S. Searches for first appearance
      of the @@ character}
    function ExtractEmailAddress(S: String; var S1, S2, EmailAddress: String): Boolean;
    { Searches all items in all units (given by field @link(Units)) for item
      S1.S2.S3 (first N  strings not empty).
      Returns a pointer to the item on success, nil otherwise. }
    function FindGlobal(const S1, S2, S3: String; N: Integer): PItem;
    function GetCIOTypeName(MyType: TCIOType): String;
    { Abstract function that provides file extension for documentation format.
      Must be overwritten by descendants. }
    function GetFileExtension: String; virtual;
    function GetSpecialCharacter(IntValue: Integer): String; virtual;
    { Returns a string constant. I must be one of the TEXT_xxx constants.
      Will call one of the language-dependent GetString methods like
      @link(GetStringEnglish), regarding the @link(Language) field. }
    function GetString(I: Integer): String;
    {@author(Rodrigo Urubatan Ferreira Jardim <rodrigo@netscape.net>)
    Gets called by @link(GetString) for Brasilian Portuguese strings to
    be used in output.}
    function GetStringBrasilian(I: Integer): String;
    { Gets called by @link(GetString) for Catalan strings to be used in
      output. }
    function GetStringCatalan(I: Integer): String;
    { Gets called by @link(GetString) for Dutch strings to be used in
      output. }
    function GetStringDutch(I: Integer): String;
    { Gets called by @link(GetString) for English strings to be used in
      output. }
    function GetStringEnglish(I: Integer): String;
    { Gets called by @link(GetString) for French strings to be used in
      output. }
    function GetStringFrench(I: Integer): String;
    { Gets called by @link(GetString) for German strings to be used in
      output. }
    function GetStringGerman(I: Integer): String;
    { Gets called by @link(GetString) for Spanish strings to be used in
      output. }
    function GetStringSpanish(I: Integer): String;
    { Loads descriptions from file N and replaces or fills the corresponding
      comment sections of items. }
    procedure LoadDescriptionFile(N: String);
    { Assumes C contains file names as PString variables.
      Calls @link(LoadDescriptionFile) with each file name. }
    procedure LoadDescriptionFiles(C: PStringCollection);
    { Searches for an item of name S which was linked in the description
      of Item.
      Starts search within item, then does a search on all items in all
      units using @link(FindGlobal).
      Returns a link as string on success or an empty string on failure. }
    function SearchLink(S: String; Item: PItem): String;
    { A link provided in a tag can be made up of up to three parts,
      separated by dots.
      If this link is not a valid identifier or if it has more than
      three parts, false is returned, true otherwise.
      The parts are returned in S1, S2 and S3, with the number of
      parts minus one being returned in N. }
    function SplitLink(S: String; var S1, S2, S3: String; var N: Integer): Boolean;
    procedure StoreDescription(ItemName: String; var T: PText);
    { Writes the names of the authors to standard output, one author per row.
      The collection Authors is expected to store zero or more names,
      one name per string.
      The heading Author(s) (@link(TEXT_AUTHOR) or @link(TEXT_AUTHORS)) is
      written in the current language at heading level HL. }
    procedure WriteAuthors(HL: Byte; Authors: PStringCollection);
    { Converts C to output format using @link(ConvertChar) and writes
      resulting string to output using @link(Stream)'s
      @link(TOutputStream.WriteString) method. }
    procedure WriteChar(C: Char);
    { Writes all information on a class, object or interface (CIO) to output,
      at heading level HL. }
    procedure WriteCIO(HL: Byte; var CIO: TCIO); virtual;
    { Writes all classes, interfaces and objects in C to output, calling
      @link(WriteCIO) with each, at heading level HL. }
    procedure WriteCIOs(HL: Byte; C: PItemCollection); virtual;
    { Abstract procedure, must be overwritten by descendants.
      Writes a list of all classes, interfaces and objects in C at heading
      level HL to output. }
    procedure WriteCIOSummary(HL: Byte; C: PItemCollection); virtual;
    { Writes collection T, which is supposed to contain constant items only
      to output at heading level HL with heading GetString(TEXT_TYPES) calling
      @link(WriteItems).
      Can be overwritten by descendants. }
    procedure WriteConstants(HL: Byte; C: PItemCollection); virtual;
    { If they are assigned, the date values for creation time and time of last
      modification are written to output at heading level HL. }
    procedure WriteDates(HL: Byte; Created, LastMod: PString); virtual;
    { Writes an already-converted description T to output.
      Takes @link(TItem.DetailedDescription) if available,
      @link(TItem.Description) otherwise.
      If none of them is assigned, nothing is written. }
    procedure WriteDescription(HL: Byte; const Heading: String; const Item: TItem);
    { Abstract procedure, must be overwritten.
      Writes all documentation.
      Will create either a single file or one file for each unit and each
      class, interface or object, depending on output format. }
    procedure WriteDocumentation; virtual;
    { Writes a list of functions / procedure or constructors / destructors /
      methods I to output.
      Heading level HL is used.
      If Methods is true, the 'Methods' heading is used, 'Functions and
      procedures' otherwise.
      Usually, a list of all items is written first, followed by detailed
      descriptions of each item.
      However, this is dependent on the output format. }
    procedure WriteFuncsProcs(HL: Byte; Methods: Boolean; FuncsProcs:
      PMethodCollection); virtual;
    { Abstract procedure that must be overwritten by descendants.
      Writes a heading S at level HL to output.
      In HTML, heading levels are regarded by choosing the appropriate
      element from H1 to H6.
      In TeX, headings will result in chapter, section, subsection or
      subsubsection elements.
      The minimum heading level is 1, the maximum level depends on the
      output format.
      However, it is no good idea to choose a heading level larger than
      five or six.
      Anyway, a descendant should be able to deal with to large HL values,
      e.g. by assigning subsubsection to all Tex headings >= 4. }
    procedure WriteHeading(HL: Byte; const S: String); virtual;
    { Writes items in I to output, including a heading of level HL and text
      Heading.
      Each item in I should be written with its short description and a
      reference.
      In HTML, this results in a table with two columns. }
    procedure WriteItems(HL: Byte; Heading: String; I: PItemCollection); virtual;
    { Calls @link(WriteString) with S, then writes a line feed. }
    procedure WriteLine(const S: String);
    { Abstract method, must be overwritten by descendants to implement
      functionality.
      Writes a list of properties P to output.
      Heading level HL is used for the heading GetString(TEXT_PROPERTIES). }      
    procedure WriteProperties(HL: Byte; P: PPropertyCollection); virtual;
    { Writes string S to output, converting each character using
      @link(ConvertChar). }
    procedure WriteString(const S: String);
    { Simply copies characters in text T to output. }
    procedure WriteText(T: PText); virtual;
    { Writes collection T, which is supposed to contain type items (TItem) to
      output at heading level HL with heading GetString(TEXT_TYPES) calling
      @link(WriteItems).
      Can be overwritten in descendants. }
    procedure WriteTypes(HL: Byte; T: PItemCollection); virtual;
    { Abstract method that writes all documentation for a single unit U to
      output, starting at heading level HL.
      Implementation must be provided by descendant objects and is dependent
      on output format.
      Will call some of the WriteXXX methods like @link(WriteHeading),
      @link(WriteCIOs) or @link(WriteUnitDescription). }
    procedure WriteUnit(HL: Byte; U: PUnit); virtual;
    { Abstract method to be implemented by descendant objects.
      Writes the (detailed, if available) description T of a unit to output,
      including a GetString(TEXT_DESCRIPTION) headline at heading level HL. }
    procedure WriteUnitDescription(HL: Byte; D: PText); virtual;
    { Writes documentation for all units, calling @link(WriteUnit) for each
      unit. }
    procedure WriteUnits(HL: Byte);
    { Writes collection V, which is supposed to contain variable items (TItem)
      to output at heading level HL with heading GetString(TEXT_TYPES) calling
      @link(WriteItems).
      Can be overwritten in descendants. }
    procedure WriteVariables(HL: Byte; V: PItemCollection); virtual;
  end;

{ Creates a string from a number, concatenating the string 'link' and a string
  version of Number. }
function AnchorToString(Number: LongInt): String;

implementation

uses
  msg;

function AnchorToString(Number: LongInt): String;
begin
  AnchorToString := 'link' + IntToString(Number);
end;

constructor TDocGenerator.Init;
begin
  inherited Init;
  Stream := nil;
end;

procedure TDocGenerator.BuildLinks;

  procedure AssignLinks(MyUnit: PUnit; MyObject: PCIO;
    const DocName: String; C: PItemCollection; var CurrentNumber: LongInt);
  var
    I: LongInt;
    P: PItem;
  begin
    if (not Assigned(C)) or (C^.Count < 1)
    then exit;
    for I := 0 to C^.Count - 1 do
    begin
      P := C^.At(I);
      P^.AnchorNumber := CurrentNumber;
      P^.MyObject := MyObject;
      P^.MyUnit := MyUnit;
      P^.FullLink := CreateLink(P);
      P^.HandleAuthorTags;
      P^.HandleCreatedTag;
      P^.HandleLastModTag;
      P^.HandleAbstractTag;
      Inc(CurrentNumber);
    end;
  end;

var
{  C: PItemCollection;}
  CO: PCIO;
  I: LongInt;
{  Item: PItem;}
  J: LongInt;
{  K: LongInt;}
  N: LongInt;
  U: PUnit;
begin
  PrintLn(3, 'Assigning links...');
  if (not Assigned(Units)) or
     (Units^.Count < 1)
  then exit;
  N:=0;
  for I := 0 to Units^.Count - 1 do
  begin
    Inc(N);
    U := Units^.At(I);
    U^.AnchorNumber:= N;
    U^.FullLink := CreateLink(U);
    U^.OutputFileName := destdir + U^.FullLink;
    U^.HandleAuthorTags;
    U^.HandleCreatedTag;
    U^.HandleLastModTag;
    U^.HandleAbstractTag;
    AssignLinks(U, nil, U^.FullLink, U^.Constants, N);
    AssignLinks(U, nil, U^.FullLink, U^.Variables, N);
    AssignLinks(U, nil, U^.FullLink, U^.Types, N);
    AssignLinks(U, nil, U^.FullLink, U^.FuncsProcs, N);
    if Assigned(U^.CIO) and (U^.CIO^.Count > 0)
    then
      begin
        for J := 0 to U^.CIO^.Count - 1 do
        begin
          Inc(N);
          CO := U^.CIO^.At(J);
          CO^.AnchorNumber:=N;
          CO^.MyUnit := U;
          CO^.OutputFileName := destdir + CO^.Name + GetFileExtension;
          CO^.FullLink := CreateLink(CO);
          CO^.HandleAuthorTags;
          CO^.HandleCreatedTag;
          CO^.HandleLastModTag;
          CO^.HandleAbstractTag;
          { Do not reset N - gives problems with TeX output}
          Inc(N);
          AssignLinks(U, CO, CO^.FullLink, CO^.Fields, N);
          AssignLinks(U, CO, CO^.FullLink, CO^.Methods, N);
          AssignLinks(U, CO, CO^.FullLink, CO^.Properties, N);
        end;
      end;
  end;
  PrintLn(3, ' ' + IntToString(N) + ' items linked.');
end;

procedure TDocGenerator.CheckForEmptyDescription(var D: PText);
var
  I: LongInt;
  NonWhitespaceFound: Boolean;
begin
  if (not Assigned(D))
  then exit;
  if (D^.Content < 1) then
  begin
    Dispose(D, Done);
    D := nil;
    exit;
  end;
  I := 0;
  repeat
    NonWhitespaceFound := not (D^.Data^[I] in Whitespace);
    Inc(I);
  until (I = D^.Content) or NonWhitespaceFound;
  if (not NonWhitespaceFound) then
  begin
    Dispose(D, Done);
    D := nil;
    exit;
  end;
end;

procedure TDocGenerator.CloseStream;
begin
  if Assigned(Stream) then
  begin
    Dispose(Stream, Done);
    Stream := nil;
  end;
end;

function TDocGenerator.ConvertChar(C: Char): String;
begin
  ConvertChar := '';
end;

function TDocGenerator.ConvertString(S: String): String;
var
  I: Integer;
  T: String;
begin
  T := '';
  I := 1;
  while (I <= Length(S)) do
  begin
    T := T + ConvertChar(S[I]);
    Inc(I);
  end;
  ConvertString := T;
end;

function TDocGenerator.CreateLink(Item: PItem): String;
begin
  CreateLink := Item^.Name;
end;

function TDocGenerator.CreateStream(Name: String): Boolean;
begin
  CloseStream;
  Name := DestDir + Name + GetFileExtension;
  PrintLn(4, 'Creating output stream "' + Name + '".');
  Stream := New(PFileOutputStream, Init(Name, true));
  CreateStream := (Stream <> nil);
end;

function TDocGenerator.CreateReferencedLink(ItemName, Link: String): String;
begin
  { abstract }
end;

function TDocGenerator.ExpandDescription(Item: PItem; var D: PText): Boolean;
var
  I: LongInt;
  J: LongInt;
{  LinkedItem: PItem;}
  Offs1: LongInt;
  Offs2: LongInt;
  Offs3: LongInt;
  S: String;
  T: PText;
begin
  ExpandDescription := true;
  { check for cases "no id" and "id is empty" }
  if (not Assigned(D))
  then exit;
  if (D^.Content < 1) then
  begin
    Dispose(D, Done);
    D := nil;
    exit;
  end;
  { create temporary TText object }
  T := New(PText, Init);
  I := 0;
  J := 0;
  repeat
    if (D^.Data^[I] = '@') then
    begin
      { this is a link or a literal @ }
      if (I < D^.Content - 1) and (D^.Data^[I + 1] = '@') then
      begin
        { literal @ }
        T^.AppendChar('@');
        Inc(I);
      end
      else { a tag }
      begin
        Offs1 := I;
        {WriteLn('Finding links in item ', Item^.Name);}
        if D^.FindTag('LINK', Offs1, Offs2, Offs3) then
        begin
          D^.GetTag(false, Offs1, Offs2, Offs3, S);
          I := Offs3 + 1;
          T^.AppendString(SearchLink(S, Item));
        end
        else
        begin
          PrintLn(2, 'Warning: Found non-link tag when ' +
            'expanding descriptions of "' + Item^.Name + '"');
          T^.AppendString('WARNING: @');
          Inc(I);
        end;
      end;
    end
    else
    begin
      if (D^.Data^[I] in [#9, #13, #10])
      then D^.Data^[I] := ' ';
      if (not KeepLineFeeds) and (J > 80) and (D^.Data^[I] in Whitespace) then
      begin
        T^.AppendString(#10);
        J := 0;
      end;
      T^.AppendString(ConvertChar(D^.Data^[I]));
      Inc(I);
    end;
    Inc(J);
  until (I >= D^.Content);
  Dispose(D, Done);
  D := T;
end;

procedure TDocGenerator.ExpandDescriptions;

  { expands Description and DetailedDescription of Item }
  procedure ExpandItem(Item: PItem);
  begin
    if (not Assigned(Item))
    then exit;
    CheckForEmptyDescription(Item^.Description);
    CheckForEmptyDescription(Item^.DetailedDescription);
    if (not ExpandDescription(Item, Item^.Description)) or
       (not ExpandDescription(Item, Item^.DetailedDescription))
    then
      begin
        PrintLn(2, 'Warning: Could not expand description from ' + Item^.Name);
      end;
  end;

  { for all items in collection C, expands descriptions }
  procedure ExpandCollection(C: PItemCollection);
  var
    I: LongInt;
    P: PItem;
    {T: PText;}
  begin
    if (not Assigned(C)) or (C^.Count < 1)
    then exit;
    for I := 0 to C^.Count - 1 do
    begin
      P := C^.At(I);
      ExpandItem(P);
    end;
  end;

var
  {C: PItemCollection;}
  CO: PCIO;
  I: LongInt;
  J: LongInt;
  U: PUnit;
begin
  PrintLn(3, 'Expanding descriptions...');
  if (not Assigned(Units)) or
     (Units^.Count < 1)
  then exit;
  for I := 0 to Units^.Count - 1 do
  begin
    U := Units^.At(I);
    ExpandItem(U);
    ExpandCollection(U^.Constants);
    ExpandCollection(U^.Variables);
    ExpandCollection(U^.Types);
    ExpandCollection(U^.FuncsProcs);
    if Assigned(U^.CIO) and (U^.CIO^.Count > 0)
    then
      begin
        for J := 0 to U^.CIO^.Count - 1 do
        begin
          CO := U^.CIO^.At(J);
          ExpandItem(CO);
          ExpandCollection(CO^.Fields);
          ExpandCollection(CO^.Methods);
          ExpandCollection(CO^.Properties);
        end;
      end;
  end;
end;

function TDocGenerator.ExtractEmailAddress(S: String; var S1, S2, EmailAddress: String): Boolean;
const
  ALLOWED_CHARS = ['a'..'z', 'A'..'Z', '-', '.', '0'..'9'];
  LETTERS = ['a'..'z', 'A'..'Z'];
var
  AtPos: Integer;
  I: Integer;
begin
  ExtractEmailAddress := false;
  if (Length(S) < 6) { minimum length of email address: a@b.cd }
  then exit;
  AtPos := Pos('@', S);
  if (AtPos < 2) or (AtPos > Length(S) - 3)
  then exit;
  { assemble address left of @ }
  I := AtPos - 1;
  while (I >= 1) and (S[I] in ALLOWED_CHARS) do
    Dec(I);
  EmailAddress := System.Copy(S, I + 1, AtPos - I - 1) + '@';
  S1 := '';
  if (I > 1)
  then S1 := System.Copy(S, 1, I);
  { assemble address right of @ }
  I := AtPos + 1;
  while (I <= Length(S)) and (S[I] in ALLOWED_CHARS) do
    Inc(I);
  EmailAddress := EmailAddress + System.Copy(S, AtPos + 1, I - AtPos - 1);
  if (Length(EmailAddress) < 6) or
     (not (EmailAddress[Length(EmailAddress)] in LETTERS)) or
     (not (EmailAddress[Length(EmailAddress) - 1] in LETTERS))
  then exit;
  S2 := '';
  if (I <= Length(S))
  then S2 := System.Copy(S, I, Length(S) - I + 1);
  ExtractEmailAddress := true;
end;

function TDocGenerator.FindGlobal(const S1, S2, S3: String; N: Integer): PItem;
var
  I: LongInt;
  Item: PItem;
  U: PUnit;
begin
  FindGlobal := nil;
  if (not Assigned(Units)) or (Units^.Count = 0)
  then exit;
  case N of
    0: { }
    for I := 0 to Units^.Count - 1 do
    begin
      U := Units^.At(I);
      Item := U^.FindItem(S1);
      if Assigned(Item) then
      begin
        FindGlobal := Item;
        exit;
      end;
    end;
    1: { unit.cio_var_const_type or object.field_method_property }
    begin
      U := PUnit(Units^.FindName(S1));
      if Assigned(U) then
      begin
        Item := U^.FindItem(S2);
        FindGlobal := Item;
        exit;
      end;
    end;
    2: { unit.objectorclassorinterface.fieldormethodorproperty }
    begin
      U := PUnit(Units^.FindName(S1));
      if (not Assigned(U))
      then exit;
{      WriteLn('DEBUG1: Found unit ' + U^.Name);}
      Item := U^.FindItem(S2);
      if (not Assigned(Item))
      then exit;
{      WriteLn('DEBUG2: Found item ' + Item^.Name);}
      Item := Item^.FindItem(S3);
      if (not Assigned(Item))
      then exit;
{      WriteLn('DEBUG3: Found item ' + Item^.Name);}
      FindGlobal := Item;
      exit;
    end;
  end;
  FindGlobal := nil;
end;

function TDocGenerator.GetCIOTypeName(MyType: TCIOType): String;
var
  S: String;
begin
  case MyType of
    CIO_CLASS: S := GetString(TEXT_CLASS);
    CIO_INTERFACE: S := GetString(TEXT_INTERFACE);
    CIO_OBJECT: S := GetString(TEXT_OBJECT);
    else S := '';
  end;
  GetCIOTypeName := S;
end;

function TDocGenerator.GetFileExtension: String;
begin
  { abstract }
  GetFileExtension := '';
end;

function TDocGenerator.GetSpecialCharacter(IntValue: Integer): String;
begin
  { abstract }
  GetSpecialCharacter := '(UNDEFINED SPECIAL CHARACTER)';
end;

function TDocGenerator.GetString(I: Integer): String;
begin
  case Language of
    LANG_ENGLISH: GetString := GetStringEnglish(I);
    LANG_DUTCH: GetString := GetStringDutch(I);
    LANG_FRENCH: GetString := GetStringFrench(I);
    LANG_GERMAN: GetString := GetStringGerman(I);
    LANG_SPANISH: GetString := GetStringSpanish(I);
    LANG_CATALAN: GetString := GetStringCatalan(I);
    else GetString :=
           'Error in TDocGenerator.GetString: Unknown output language.';
  end;
end;

function TDocGenerator.GetStringBrasilian(I: Integer): String;
begin
  case I of
    TEXT_AUTHOR: GetStringBrasilian := 'Autor';
    TEXT_AUTHORS: GetStringBrasilian := 'Autores';
    TEXT_CIO_SUMMARY: GetStringBrasilian := 'Lista de classes, interfaces e objetos';
    TEXT_CLASS: GetStringBrasilian := 'Classe';
    TEXT_CLASSES: GetStringBrasilian := 'Classes';
    TEXT_CIO: GetStringBrasilian := 'Classes, interfaces e objetos';
    TEXT_CONSTANTS: GetStringBrasilian := 'Constantes';
    TEXT_CREATED: GetStringBrasilian := 'Criado';
    TEXT_DESCRIPTION: GetStringBrasilian := 'Descrio';
    TEXT_FIELDS: GetStringBrasilian := 'Campos';
    TEXT_FUNCTIONS_AND_PROCEDURES: GetStringBrasilian := 'Funes e Procedimentos';
    TEXT_HEADLINE_UNITS: GetStringBrasilian := 'Todos os Fontes';
    TEXT_HEADLINE_CIO: GetStringBrasilian := 'Todas as classes, interfaces e objetos';
    TEXT_HEADLINE_TYPES: GetStringBrasilian := 'Todos os tipos';
    TEXT_HEADLINE_VARIABLES: GetStringBrasilian := 'Todas as variveis';
    TEXT_HEADLINE_CONSTANTS: GetStringBrasilian := 'Todas as constantes';
    TEXT_HEADLINE_FUNCSPROCS: GetStringBrasilian := 'Todas as funes e procedimentos';
    TEXT_HEADLINE_IDENTIFIERS: GetStringBrasilian := 'Todos os identificadores';
    TEXT_INTERFACE: GetStringBrasilian := 'Interface';
    TEXT_LASTMOD: GetStringBrasilian := 'Ultima modificao';
    TEXT_METHODS: GetStringBrasilian := 'Metodos';
    TEXT_NAME: GetStringBrasilian := 'Nome';
    TEXT_NONE: GetStringBrasilian := 'Nada.';
    TEXT_OBJECT: GetStringBrasilian := 'Objeto';
    TEXT_OBJECTS: GetStringBrasilian := 'Objetos';
    TEXT_OVERVIEW: GetStringBrasilian := 'Overview';
    TEXT_OVERVIEW_UNITS: GetStringBrasilian := 'FONTES';
    TEXT_OVERVIEW_CIO: GetStringBrasilian := 'CLASSES, INTERFACES, OBJETOS';
    TEXT_OVERVIEW_TYPES: GetStringBrasilian := 'TIPOS';
    TEXT_OVERVIEW_VARIABLES: GetStringBrasilian := 'VARIVEIS';
    TEXT_OVERVIEW_CONSTANTS: GetStringBrasilian := 'CONSTANTES';
    TEXT_OVERVIEW_FUNCSPROCS: GetStringBrasilian := 'FUNES, PROCEDIMENTOS';
    TEXT_OVERVIEW_IDENTIFIERS: GetStringBrasilian := 'IDENTIFICADORES';
    TEXT_PROPERTIES: GetStringBrasilian := 'Propriedades';
    TEXT_SUMMARY_OF_PROPERTIES: GetStringBrasilian := 'Lista de propriedades';
    TEXT_TYPES: GetStringBrasilian := 'Tipos';
    TEXT_UNIT: GetStringBrasilian := 'Fonte';
    TEXT_VARIABLES: GetStringBrasilian := 'Variaveis';
    TEXT_WARNING_AUTOMATIC: GetStringBrasilian :=
	     'Aviso, no altere - este arquivo foi criado automaticamente e sera apagado.';
    TEXT_SUMMARY_OF_METHODS: GetStringBrasilian := 'Lista de metodos';
    TEXT_DECLARATION: GetStringBrasilian := 'Declarao';
    TEXT_TYPE: GetStringBrasilian := 'Tipo';
    else GetStringBrasilian := 'Erro em TDocGenerator.GetBrasilianString - string (frase) desconhecida.';
  end;
end;

function TDocGenerator.GetStringCatalan(I: Integer): String;
begin
  case I of
    TEXT_AUTHOR: GetStringCatalan := 'Autor';
    TEXT_AUTHORS: GetStringCatalan := 'Autors';
    TEXT_CIO_SUMMARY: GetStringCatalan := 'Llista de clases, interfaces i objectes';
    TEXT_CLASS: GetStringCatalan := 'Clase';
    TEXT_CLASSES: GetStringCatalan := 'Clases';
    TEXT_CIO: GetStringCatalan := 'Clases, interfaces i objectes';
    TEXT_CONSTANTS: GetStringCatalan := 'Constants';
    TEXT_CREATED: GetStringCatalan := 'Creat';
    TEXT_DESCRIPTION: GetStringCatalan := 'Descripci';
    TEXT_FIELDS: GetStringCatalan := 'Camps';
    TEXT_FUNCTIONS_AND_PROCEDURES: GetStringCatalan := 'Funcions i procediments';
    TEXT_HEADLINE_UNITS: GetStringCatalan := 'Totes les unitats';
    TEXT_HEADLINE_CIO: GetStringCatalan := 'Totes les clases, interfaces i objectes';
    TEXT_HEADLINE_TYPES: GetStringCatalan := 'Tots els tipus';
    TEXT_HEADLINE_VARIABLES: GetStringCatalan := 'Totes les variables';
    TEXT_HEADLINE_CONSTANTS: GetStringCatalan := 'Totes les constants';
    TEXT_HEADLINE_FUNCSPROCS: GetStringCatalan := 'Totes les funcions i procediments';
    TEXT_HEADLINE_IDENTIFIERS: GetStringCatalan := 'Tot els indentificadors';
    TEXT_INTERFACE: GetStringCatalan := 'Interface';
    TEXT_LASTMOD: GetStringCatalan := 'ltima modificaci';
    TEXT_METHODS: GetStringCatalan := 'Mtodes';
    TEXT_NAME: GetStringCatalan := 'Nom';
    TEXT_NONE: GetStringCatalan := 'Ningu.';
    TEXT_OBJECT: GetStringCatalan := 'Objecte';
    TEXT_OBJECTS: GetStringCatalan := 'Objectes';
    TEXT_OVERVIEW: GetStringCatalan := 'Resum';
    TEXT_OVERVIEW_UNITS: GetStringCatalan := 'UNITATS';
    TEXT_OVERVIEW_CIO: GetStringCatalan := 'CLASES, INTERFACES, OBJECTES';
    TEXT_OVERVIEW_TYPES: GetStringCatalan := 'TIPUS';
    TEXT_OVERVIEW_VARIABLES: GetStringCatalan := 'VARIABLES';
    TEXT_OVERVIEW_CONSTANTS: GetStringCatalan := 'CONSTANTS';
    TEXT_OVERVIEW_FUNCSPROCS: GetStringCatalan := 'FUNCIONS, PROCEDIMENTS';
    TEXT_OVERVIEW_IDENTIFIERS: GetStringCatalan := 'IDENTIFICADORS';
    TEXT_PROPERTIES: GetStringCatalan := 'Propietats';
    TEXT_SUMMARY_OF_PROPERTIES: GetStringCatalan := 'Llista de propietats';
    TEXT_TYPES: GetStringCatalan := 'Tipus';
    TEXT_UNIT: GetStringCatalan := 'Unitat';
    TEXT_VARIABLES: GetStringCatalan := 'Variables';
    TEXT_WARNING_AUTOMATIC: GetStringCatalan :=
      'Atenci, no editar - aquest fitxer ha estat creat automaticament i ser sobrescrit';
    TEXT_SUMMARY_OF_METHODS: GetStringCatalan := 'Llista de mtodes';
    TEXT_DECLARATION: GetStringCatalan := 'Declaraci';
    TEXT_TYPE: GetStringCatalan := 'Tipus';
    else GetStringCatalan := 'Error en TDocGenerator.GetEnglishString - cadena desconeguda.';
  end;
end;

function TDocGenerator.GetStringDutch(I: Integer): String;
begin
  case I of
    TEXT_AUTHOR: GetStringDutch := 'Auteur';
    TEXT_AUTHORS: GetStringDutch := 'Auteurs';
    TEXT_CIO_SUMMARY: GetStringDutch := 'Overzicht van classes, interfaces & objecten';
    TEXT_CLASS: GetStringDutch := 'Class';
    TEXT_CLASSES: GetStringDutch := 'Classes';
    TEXT_CIO: GetStringDutch := 'Classes, interfaces and objecten';
    TEXT_CONSTANTS: GetStringDutch := 'Constanten';
    TEXT_CREATED: GetStringDutch := 'Gemaakt';
    TEXT_DESCRIPTION: GetStringDutch := 'Omschrijving';
    TEXT_FIELDS: GetStringDutch := 'Velden';
    TEXT_FUNCTIONS_AND_PROCEDURES: GetStringDutch := 'Functies en procedures';
    TEXT_HEADLINE_UNITS: GetStringDutch := 'Alle units';
    TEXT_HEADLINE_CIO: GetStringDutch := 'Alle classes, interfaces en objecten';
    TEXT_HEADLINE_TYPES: GetStringDutch := 'Alle typen';
    TEXT_HEADLINE_VARIABLES: GetStringDutch := 'Alle variabelen';
    TEXT_HEADLINE_CONSTANTS: GetStringDutch := 'Alle constanten';
    TEXT_HEADLINE_FUNCSPROCS: GetStringDutch := 'Alle functies en procedures';
    TEXT_HEADLINE_IDENTIFIERS: GetStringDutch := 'Alle identifiers';
    TEXT_INTERFACE: GetStringDutch := 'Interface';
    TEXT_LASTMOD: GetStringDutch := 'Laatste wijziging';
    TEXT_METHODS: GetStringDutch := 'Methods';
    TEXT_NAME: GetStringDutch := 'Naam';
    TEXT_NONE: GetStringDutch := 'Geen.';
    TEXT_OBJECT: GetStringDutch := 'Object';
    TEXT_OBJECTS: GetStringDutch := 'Objecten';
    TEXT_OVERVIEW: GetStringDutch := 'Overzicht';
    TEXT_OVERVIEW_UNITS: GetStringDutch := 'UNITS';
    TEXT_OVERVIEW_CIO: GetStringDutch := 'CLASSES, INTERFACES, OBJECTEN';
    TEXT_OVERVIEW_TYPES: GetStringDutch := 'TYPEN';
    TEXT_OVERVIEW_VARIABLES: GetStringDutch := 'VARIABELEN';
    TEXT_OVERVIEW_CONSTANTS: GetStringDutch := 'CONSTANTEN';
    TEXT_OVERVIEW_FUNCSPROCS: GetStringDutch := 'FUNCTIES, PROCEDURES';
    TEXT_OVERVIEW_IDENTIFIERS: GetStringDutch := 'IDENTIFIERS';
    TEXT_PROPERTIES: GetStringDutch := 'Eigenschappen';
    TEXT_SUMMARY_OF_PROPERTIES: GetStringDutch := 'Overzicht van eigenschappen';
    TEXT_TYPES: GetStringDutch := 'Typen';
    TEXT_UNIT: GetStringDutch := 'Unit';
    TEXT_VARIABLES: GetStringDutch := 'Variabelen';
    TEXT_WARNING_AUTOMATIC: GetStringDutch := 'Waarschuwing, wijzig niets - dit bestand is automatisch gegenereerd en zal worden overschreven';
    TEXT_SUMMARY_OF_METHODS: GetStringDutch := 'Overzicht van methods';
    TEXT_DECLARATION: GetStringDutch := 'Declaratie';
    TEXT_TYPE: GetStringDutch := 'Type';
    else GetStringDutch := 'Er is een fout opgetreden in TDocGenerator.GetDutchString - onbekende text.';
   end;
end;

function TDocGenerator.GetStringEnglish(I: Integer): String;
begin
  case I of
    TEXT_AUTHOR: GetStringEnglish := 'Author';
    TEXT_AUTHORS: GetStringEnglish := 'Authors';
    TEXT_CIO_SUMMARY: GetStringEnglish := 'Summary of classes, interfaces & objects';
    TEXT_CLASS: GetStringEnglish := 'Class';
    TEXT_CLASSES: GetStringEnglish := 'Classes';
    TEXT_CIO: GetStringEnglish := 'Classes, interfaces and objects';
    TEXT_CONSTANTS: GetStringEnglish := 'Constants';
    TEXT_CREATED: GetStringEnglish := 'Created';
    TEXT_DESCRIPTION: GetStringEnglish := 'Description';
    TEXT_FIELDS: GetStringEnglish := 'Fields';
    TEXT_FUNCTIONS_AND_PROCEDURES: GetStringEnglish := 'Functions and procedures';
    TEXT_HEADLINE_UNITS: GetStringEnglish := 'All units';
    TEXT_HEADLINE_CIO: GetStringEnglish := 'All classes, interfaces and objects';
    TEXT_HEADLINE_TYPES: GetStringEnglish := 'All types';
    TEXT_HEADLINE_VARIABLES: GetStringEnglish := 'All variables';
    TEXT_HEADLINE_CONSTANTS: GetStringEnglish := 'All constants';
    TEXT_HEADLINE_FUNCSPROCS: GetStringEnglish := 'All functions and procedures';
    TEXT_HEADLINE_IDENTIFIERS: GetStringEnglish := 'All identifiers';
    TEXT_INTERFACE: GetStringEnglish := 'Interface';
    TEXT_LASTMOD: GetStringEnglish := 'Last modification';
    TEXT_METHODS: GetStringEnglish := 'Methods';
    TEXT_NAME: GetStringEnglish := 'Name';
    TEXT_NONE: GetStringEnglish := 'None.';
    TEXT_OBJECT: GetStringEnglish := 'Object';
    TEXT_OBJECTS: GetStringEnglish := 'Objects';
    TEXT_OVERVIEW: GetStringEnglish := 'Overview';
    TEXT_OVERVIEW_UNITS: GetStringEnglish := 'UNITS';
    TEXT_OVERVIEW_CIO: GetStringEnglish := 'CLASSES, INTERFACES, OBJECTS';
    TEXT_OVERVIEW_TYPES: GetStringEnglish := 'TYPES';
    TEXT_OVERVIEW_VARIABLES: GetStringEnglish := 'VARIABLES';
    TEXT_OVERVIEW_CONSTANTS: GetStringEnglish := 'CONSTANTS';
    TEXT_OVERVIEW_FUNCSPROCS: GetStringEnglish := 'FUNCTIONS, PROCEDURES';
    TEXT_OVERVIEW_IDENTIFIERS: GetStringEnglish := 'IDENTIFIERS';
    TEXT_PROPERTIES: GetStringEnglish := 'Properties';
    TEXT_SUMMARY_OF_PROPERTIES: GetStringEnglish := 'Summary of properties';
    TEXT_TYPES: GetStringEnglish := 'Types';
    TEXT_UNIT: GetStringEnglish := 'Unit';
    TEXT_VARIABLES: GetStringEnglish := 'Variables';
    TEXT_WARNING_AUTOMATIC: GetStringEnglish :=
      'Warning, do not edit - this file has been created automatically and will be overwritten';
    TEXT_SUMMARY_OF_METHODS: GetStringEnglish := 'Summary of methods';
    TEXT_DECLARATION: GetStringEnglish := 'Declaration';
    TEXT_TYPE: GetStringEnglish := 'Type';
    else GetStringEnglish := 'Error in TDocGenerator.GetEnglishString - unknown string.';
  end;
end;

function TDocGenerator.GetStringFrench(I: Integer): String;
begin
  case I of
    TEXT_AUTHOR: GetStringFrench := 'Auteur';
    TEXT_AUTHORS: GetStringFrench := 'Auteurs';
    TEXT_CIO_SUMMARY: GetStringFrench := 'Classes, interfaces & objets';
    TEXT_CLASS: GetStringFrench := 'Classe';
    TEXT_CLASSES: GetStringFrench := 'Classes';
    TEXT_CIO: GetStringFrench := 'Classes, interfaces and objects';
    TEXT_CONSTANTS: GetStringFrench := 'Constantes';
    TEXT_CREATED: GetStringFrench := 'Cre';
    TEXT_DESCRIPTION: GetStringFrench := 'Description';
    TEXT_FIELDS: GetStringFrench := 'Champs';
    TEXT_FUNCTIONS_AND_PROCEDURES: GetStringFrench := 'Fonctions et procdures';
    TEXT_HEADLINE_UNITS: GetStringFrench := 'Toutes les units';
    TEXT_HEADLINE_CIO: GetStringFrench := 'Toutes les classes, interfaces et objets';
    TEXT_HEADLINE_TYPES: GetStringFrench := 'Tous les types';
    TEXT_HEADLINE_VARIABLES: GetStringFrench := 'Toutes les variables';
    TEXT_HEADLINE_CONSTANTS: GetStringFrench := 'Toutes les constants';
    TEXT_HEADLINE_FUNCSPROCS: GetStringFrench := 'Toutes les fonctions et procdures';
    TEXT_HEADLINE_IDENTIFIERS: GetStringFrench := 'Tous les identificateurs';
    TEXT_INTERFACE: GetStringFrench := 'Interface';
    TEXT_LASTMOD: GetStringFrench := 'Dernire modification';
    TEXT_METHODS: GetStringFrench := 'Mthodes';
    TEXT_NAME: GetStringFrench := 'Nom';
    TEXT_NONE: GetStringFrench := 'Aucun(e)(s).';
    TEXT_OBJECT: GetStringFrench := 'Objet';
    TEXT_OBJECTS: GetStringFrench := 'Objets';
    TEXT_OVERVIEW: GetStringFrench := 'Aperu';
    TEXT_OVERVIEW_UNITS: GetStringFrench := 'UNITS';
    TEXT_OVERVIEW_CIO: GetStringFrench := 'CLASSES, INTERFACES, OBJETS';
    TEXT_OVERVIEW_TYPES: GetStringFrench := 'TYPES';
    TEXT_OVERVIEW_VARIABLES: GetStringFrench := 'VARIABLES';
    TEXT_OVERVIEW_CONSTANTS: GetStringFrench := 'CONSTANTES';
    TEXT_OVERVIEW_FUNCSPROCS: GetStringFrench := 'FONCTIONS, PROCDURES';
    TEXT_OVERVIEW_IDENTIFIERS: GetStringFrench := 'IDENTIFICATEURS';
    TEXT_PROPERTIES: GetStringFrench := 'Proprits';
    TEXT_SUMMARY_OF_PROPERTIES: GetStringFrench := 'Aperu des proprits';
    TEXT_TYPES: GetStringFrench := 'Types';
    TEXT_UNIT: GetStringFrench := 'Unit';
    TEXT_VARIABLES: GetStringFrench := 'Variables';
    TEXT_WARNING_AUTOMATIC:
      GetStringFrench := 'Attention, ne pas dtier - ce fichier est cr automatiquement et va tre cras';
    TEXT_SUMMARY_OF_METHODS: GetStringFrench := 'Aperu des mthods';
    TEXT_DECLARATION: GetStringFrench := 'Dclaration';
    TEXT_TYPE: GetStringFrench := 'Type';
    else GetStringFrench := 'Erreur dans TDocGenerator.GetFrenchString - chane inconnue.';
  end;
end;

function TDocGenerator.GetStringGerman(I: Integer): String;
begin
  case I of
    TEXT_AUTHOR: GetStringGerman := 'Autor';
    TEXT_AUTHORS: GetStringGerman := 'Autoren';
    TEXT_CIO_SUMMARY: GetStringGerman := 'Liste aller Klassen, Schnittstellen und Objekte';
    TEXT_CLASS: GetStringGerman := 'Klasse';
    TEXT_CLASSES: GetStringGerman := 'Klassen';
    TEXT_CIO: GetStringGerman := 'Klassen, Schnittstellen und Objekte';
    TEXT_CONSTANTS: GetStringGerman := 'Konstanten';
    TEXT_CREATED: GetStringGerman := 'Erstellt';
    TEXT_DESCRIPTION: GetStringGerman := 'Beschreibung';
    TEXT_FIELDS: GetStringGerman := 'Felder';
    TEXT_FUNCTIONS_AND_PROCEDURES: GetStringGerman := 'Funktionen und Prozeduren';
    TEXT_HEADLINE_UNITS: GetStringGerman := 'Alle Units';
    TEXT_HEADLINE_CIO: GetStringGerman := 'Alle Klassen, Schnittstellen und Objekte';
    TEXT_HEADLINE_TYPES: GetStringGerman := 'Alle Typen';
    TEXT_HEADLINE_VARIABLES: GetStringGerman := 'Alle Variablen';
    TEXT_HEADLINE_CONSTANTS: GetStringGerman := 'Alle Konstanten';
    TEXT_HEADLINE_FUNCSPROCS: GetStringGerman := 'Alle Funktionen und Prozeduren';
    TEXT_HEADLINE_IDENTIFIERS: GetStringGerman := 'Alle Bezeichner';
    TEXT_INTERFACE: GetStringGerman := 'Schnittstelle';
    TEXT_LASTMOD: GetStringGerman := 'Letzte Vernderung';
    TEXT_METHODS: GetStringGerman := 'Methoden';
    TEXT_NAME: GetStringGerman := 'Name';
    TEXT_NONE: GetStringGerman := 'Keine.';
    TEXT_OBJECT: GetStringGerman := 'Objekt';
    TEXT_OBJECTS: GetStringGerman := 'Objekte';
    TEXT_OVERVIEW: GetStringGerman := 'berblick';
    TEXT_OVERVIEW_UNITS: GetStringGerman := 'UNITS';
    TEXT_OVERVIEW_CIO: GetStringGerman := 'KLASSEN, SCHNITTSTELLEN, OBJEKTE';
    TEXT_OVERVIEW_TYPES: GetStringGerman := 'TYPEN';
    TEXT_OVERVIEW_VARIABLES: GetStringGerman := 'VARIABLEN';
    TEXT_OVERVIEW_CONSTANTS: GetStringGerman := 'KONSTANTEN';
    TEXT_OVERVIEW_FUNCSPROCS: GetStringGerman := 'FUNKTIONEN, PROZEDUREN';
    TEXT_OVERVIEW_IDENTIFIERS: GetStringGerman := 'BEZEICHNER';
    TEXT_PROPERTIES: GetStringGerman := 'Eigenschaften';
    TEXT_SUMMARY_OF_PROPERTIES: GetStringGerman := 'Liste aller Eigenschaften';
    TEXT_TYPES: GetStringGerman := 'Typen';
    TEXT_UNIT: GetStringGerman := 'Unit';
    TEXT_VARIABLES: GetStringGerman := 'Variablen';
    TEXT_WARNING_AUTOMATIC: GetStringGerman :=
    'Warnung, nicht editieren - diese Datei wurde automatisch erstellt und wird berschrieben werden.';
    TEXT_SUMMARY_OF_METHODS: GetStringGerman := 'Liste aller Methoden';
    TEXT_DECLARATION: GetStringGerman := 'Deklaration';
    TEXT_TYPE: GetStringGerman := 'Typ';
    else GetStringGerman := 'Fehler in TDocGenerator.GetEnglishString - unbekannter String.';
  end;
end;

function TDocGenerator.GetStringSpanish(I: Integer): String;
begin
  case I of
    TEXT_AUTHOR: GetStringSpanish := 'Autor';
    TEXT_AUTHORS: GetStringSpanish := 'Autores';
    TEXT_CIO_SUMMARY: GetStringSpanish := 'Lista de clases, interfaces y objetos';
    TEXT_CLASS: GetStringSpanish := 'Clase';
    TEXT_CLASSES: GetStringSpanish := 'Clases';
    TEXT_CIO: GetStringSpanish := 'Clases, interfaces y objetos';
    TEXT_CONSTANTS: GetStringSpanish := 'Constantes';
    TEXT_CREATED: GetStringSpanish := 'Creado';
    TEXT_DESCRIPTION: GetStringSpanish := 'Descripcion';
    TEXT_FIELDS: GetStringSpanish := 'Campos';
    TEXT_FUNCTIONS_AND_PROCEDURES: GetStringSpanish := 'Funciones y procedimientos';
    TEXT_HEADLINE_UNITS: GetStringSpanish := 'Todas las unidades';
    TEXT_HEADLINE_CIO: GetStringSpanish := 'Todas las clases, interfaces y objetos';
    TEXT_HEADLINE_TYPES: GetStringSpanish := 'Todos los tipos';
    TEXT_HEADLINE_VARIABLES: GetStringSpanish := 'Todas las variables';
    TEXT_HEADLINE_CONSTANTS: GetStringSpanish := 'Todas las constantes';
    TEXT_HEADLINE_FUNCSPROCS: GetStringSpanish := 'Todos las funciones y procedimientos';
    TEXT_HEADLINE_IDENTIFIERS: GetStringSpanish := 'Todos los indentificadores';
    TEXT_INTERFACE: GetStringSpanish := 'Interface';
    TEXT_LASTMOD: GetStringSpanish := 'ltima modificacin';
    TEXT_METHODS: GetStringSpanish := 'Mtodos';
    TEXT_NAME: GetStringSpanish := 'Nombre';
    TEXT_NONE: GetStringSpanish := 'Ninguno.';
    TEXT_OBJECT: GetStringSpanish := 'Objeto';
    TEXT_OBJECTS: GetStringSpanish := 'Objetos';
    TEXT_OVERVIEW: GetStringSpanish := 'Resumen';
    TEXT_OVERVIEW_UNITS: GetStringSpanish := 'UNIDADES';
    TEXT_OVERVIEW_CIO: GetStringSpanish := 'CLASES, INTERFACES, OBJETOS';
    TEXT_OVERVIEW_TYPES: GetStringSpanish := 'TIPOS';
    TEXT_OVERVIEW_VARIABLES: GetStringSpanish := 'VARIABLES';
    TEXT_OVERVIEW_CONSTANTS: GetStringSpanish := 'CONSTANTES';
    TEXT_OVERVIEW_FUNCSPROCS: GetStringSpanish := 'FUNCIONES, PROCEDIMIENTOS';
    TEXT_OVERVIEW_IDENTIFIERS: GetStringSpanish := 'IDENTIFICADORES';
    TEXT_PROPERTIES: GetStringSpanish := 'Propiedades';
    TEXT_SUMMARY_OF_PROPERTIES: GetStringSpanish := 'Lista de propiedades';
    TEXT_TYPES: GetStringSpanish := 'Tipos';
    TEXT_UNIT: GetStringSpanish := 'Unidad';
    TEXT_VARIABLES: GetStringSpanish := 'Variables';
    TEXT_WARNING_AUTOMATIC: GetStringSpanish :=
      'Atencin, no editar - este fichero ha sido creado automaticamente y ser sobrescrito';
    TEXT_SUMMARY_OF_METHODS: GetStringSpanish := 'Lista de mtodos';
    TEXT_DECLARATION: GetStringSpanish := 'Declaracin';
    TEXT_TYPE: GetStringSpanish := 'Tipos';
    else GetStringSpanish := 'Error en TDocGenerator.GetStringSpanish - cadena desconocida.';
  end;
end;

procedure TDocGenerator.LoadDescriptionFile(N: String);
var
  F: PFileInputStream;
  I: PItem;
  ItemName: String;
  S: String;
  T: PText;
begin
  if (Length(N) < 1)
  then exit;
  PrintLn(3, 'Loading descriptions from file "' + N + '"');
  F := New(PFileInputStream, Init(N));
  if (not Assigned(F)) then
  begin
    PrintLn(2, 'Error: Could not open description file "' + N + '"');
    exit;
  end;
  I := nil;
  T := nil;
  while (F^.HasData) do
  begin
    F^.ReadLine(S);
{    PrintLn(4, 'DEBUG - descr "' + S + '"');}
    if (Length(S) > 0) then
    begin
      { # means: description of another item begins }
      if (S[1] = '#') then
      begin
      	{ if there is an old description, deal with it }
      	StoreDescription(ItemName, T);
        { delete # char }
        System.Delete(S, 1, 1);
        { skip whitespace }
        while (Length(S) > 0) and (S[1] in [' ', #9]) do
          System.Delete(S, 1, 1);
        { find item }
        ItemName := '';
        while (Length(S) > 0) and
          (S[1] in ['A'..'Z', 'a'..'z', '_', '.', '0'..'9']) do
        begin
          ItemName := ItemName + S[1];
          System.Delete(S, 1, 1);
        end;
        T := New(PText, Init);
        {PrintLn(2, 'DEBUG - loaded description for "' + ItemName + '"');}
        { dispose of item's current description }
        {if Assigned(I) and Assigned(I^.Description) then
        begin
          Dispose(I^.Description, Done);
          T := New(PText, Init);
          if (not Assigned(T)) then
          begin
          end;
          I^.Description := T;
        end;}
      end
      else
      begin
        { check if there is a text }
        if (not Assigned(T)) then
        begin
          PrintLn(2, 'Error: First line of description file must start ' +
            'with "# item_name"');
          Break; { leave while loop }
        end;
        T^.AppendString(S + #10);
      end;
    end;
  end;
  StoreDescription(ItemName, T);
  Dispose(F, Done);
end;

procedure TDocGenerator.LoadDescriptionFiles(C: PStringCollection);
var
  I: LongInt;
  S: PString;
begin
  if (not Assigned(C)) or (C^.Count < 1)
  then exit;
  PrintLn(3, 'Loading description files...');
  for I := 0 to C^.Count - 1 do
  begin
    S := PString(C^.At(I));
    if Assigned(S)
    then LoadDescriptionFile(S^);
  end;
end;

function TDocGenerator.SearchLink(S: String; Item: PItem): String;
var
  N: Integer;
  S1: String;
  S2: String;
  S3: String;
  FoundItem: PItem;
begin
  { S is supposed to have 0 to 2 dots in it - S1, S2 and S3 contain
      the parts between the dots, N the number of dots }
  if (not SplitLink(S, S1, S2, S3, N)) then
  begin
    PrintLn(2, 'Warning: the link "' + S + '" is invalid');
    SearchLink := 'UNKNOWN';
    exit;
  end;
  { first try to find link starting at Item }
  if Assigned(Item) then
  begin
    FoundItem := Item^.FindName(S1, S2, S3, N);
  end
  else FoundItem := nil;
  if (not Assigned(FoundItem))
  then FoundItem := FindGlobal(S1, S2, S3, N);
  if Assigned(FoundItem)
  then SearchLink := CreateReferencedLink(FoundItem^.Name, FoundItem^.FullLink)
  else SearchLink := ConvertString(S);
end;

function TDocGenerator.SplitLink(S: String; var S1, S2, S3: String;
  var N: Integer): Boolean;

  function RemoveLeadingAndTrailingWhitespace(S: String): String;
  begin
    while (Length(S) > 0) and (S[1] in Whitespace) do
      System.Delete(S, 1, 1);
    while (Length(S) > 0) and (S[Length(S)] in Whitespace) do
      System.Delete(S, Length(S), 1);
    RemoveLeadingAndTrailingWhitespace := S;
  end;

  procedure SplitInTwo(S: String; var S1, S2: String);
  var
    I: Integer;
  begin
    I := Pos('.', S);
    if (I = 0)
    then
      begin
        S1 := S;
        S2 := '';
      end
    else
      begin
        S1 := System.Copy(S, 1, I - 1);
        S2 := System.Copy(S, I + 1, Length(S));
      end;
  end;

var
  I: LongInt;
  T: String;
begin
  SplitLink := false;
  S1 := '';
  S2 := '';
  S3 := '';
  N := 0;
{  I := 1;}
  S := RemoveLeadingAndTrailingWhitespace(S);
  if (Length(S) = 0)
  then exit;
  if (not (S[1] in IdentifierStart))
  then exit;
  I := 2;
  while (I <= Length(S)) do
  begin
    if (not (S[I] in IdentifierOther))
    then exit;
    Inc(I);
  end;
  SplitInTwo(S, S1, S2);
  if (Length(S2) = 0)
  then
    begin
      N := 0;
    end
  else
    begin
      T := S2;
      SplitInTwo(T, S2, S3);
      if (Length(S3) = 0)
      then N := 1
      else N := 2;
    end;
  SplitLink := true;
end;

procedure TDocGenerator.StoreDescription(ItemName: String; var T: PText);
var
  Item: PItem;
  N: Integer;
  S1: String;
  S2: String;
  S3: String;
begin
  if (not Assigned(T)) then
  begin
  	exit;
  end;
  PrintLn(5, 'Storing description for ' + ItemName);
  if SplitLink(ItemName, S1, S2, S3, N) then
  begin
{  	Write('DEBUG Split into ', N + 1, ' pieces: ');
  	if (N >= 0) then Write(S1 + ' ');
  	if (N >= 1) then Write(S2 + ' ');
  	if (N >= 2) then Write(S3 + ' ');
  	WriteLn;}
    Item := FindGlobal(S1, S2, S3, N);
    if Assigned(Item) then
    begin
      if Assigned(Item^.Description) then
      begin
        PrintLn(2, 'Warning - more than one description for ' + ItemName);
        Dispose(T, Done);
      end
      else
      begin
      	Item^.Description := T;
      end;
    end
    else
    begin
      PrintLn(2, 'Warning - could not find item ' + ItemName);
      Dispose(T, Done);
    end;
  end
  else
  begin
  	PrintLn(2, 'Warning: Could not split item "' + ItemName + '"');
  end;
  T := nil;
end;

procedure TDocGenerator.WriteAuthors(HL: Byte; Authors: PStringCollection);
begin
  { abstract }
end;

procedure TDocGenerator.WriteChar(C: Char);
begin
  Stream^.WriteString(ConvertChar(C));
end;

procedure TDocGenerator.WriteCIO(HL: Byte; var CIO: TCIO);
begin
  { abstract }
end;

procedure TDocGenerator.WriteCIOs(HL: Byte; C: PItemCollection);
var
  I: LongInt;
  P: PCIO;
begin
  if (not Assigned(C)) or (C^.Count < 1)
  then exit;
  for I := 0 to C^.Count - 1 do
  begin
    P := C^.At(I);
    WriteCIO(HL, P^);
  end;
end;

procedure TDocGenerator.WriteCIOSummary(HL: Byte; C: PItemCollection);
begin
  WriteItems(HL, GetString(TEXT_CIO_SUMMARY), C);
end;

procedure TDocGenerator.WriteConstants(HL: Byte; C: PItemCollection);
begin
  WriteItems(HL, GetString(TEXT_CONSTANTS), C);
end;

procedure TDocGenerator.WriteDates(HL: Byte; Created, LastMod: PString);
begin
  { abstract }
end;

procedure TDocGenerator.WriteDescription(HL: Byte; const Heading: String; const Item: TItem);
var
  D: PText;
begin
  if (Assigned(Item.DetailedDescription))
  then D := Item.DetailedDescription
  else
    begin
      if (Assigned(Item.Description))
      then D := Item.Description
      else exit;
    end;
  if (Length(Heading) > 0)
  then WriteHeading(HL, Heading);
  WriteText(D);
end;

procedure TDocGenerator.WriteDocumentation;
begin
  { abstract }
end;

procedure TDocGenerator.WriteFuncsProcs(HL: Byte; Methods: Boolean; FuncsProcs: PMethodCollection);
begin
  { abstract }
end;

procedure TDocGenerator.WriteHeading(HL: Byte; const S: String);
begin
  { abstract }
end;

procedure TDocGenerator.WriteItems(HL: Byte; Heading: String; I: PItemCollection);
begin
  { abstract }
end;

procedure TDocGenerator.WriteLine(const S: String);
begin
  WriteString(S);
  Stream^.WriteLine('');
end;

procedure TDocGenerator.WriteProperties(HL: Byte; P: PPropertyCollection);
begin
  { abstract }
end;

procedure TDocGenerator.WriteString(const S: String);
var
  I: LongInt;
  L: LongInt;
  T: String;
begin
  L := Length(S);
  if (L = 0)
  then exit;
  T := '';
  for I := 1 to L do
    T := T + ConvertChar(S[I]);
  Stream^.WriteString(T);
end;

procedure TDocGenerator.WriteText(T: PText);
var
  I: Integer;
begin
  if (T = nil) or (T^.Content < 1)
  then exit;
  I := 0;
  while (I < T^.Content) do
  begin
    Stream^.WriteChar(T^.Data^[I]);
    Inc(I);
  end;
end;

procedure TDocGenerator.WriteTypes(HL: Byte; T: PItemCollection);
begin
  WriteItems(HL, GetString(TEXT_TYPES), T);
end;

procedure TDocGenerator.WriteUnit(HL: Byte; U: PUnit);
begin
  { abstract }
end;

procedure TDocGenerator.WriteUnitDescription(HL: Byte; D: PText);
begin
  { abstract }
end;

procedure TDocGenerator.WriteUnits(HL: Byte);
var
  I: LongInt;
  P: PUnit;
begin
  if (not Assigned(Units)) or (Units^.Count < 1)
  then exit;
  for I := 0 to Units^.Count - 1 do
  begin
    P := PUnit(Units^.At(I));
    WriteUnit(HL, P);
  end;
end;

procedure TDocGenerator.WriteVariables(HL: Byte; V: PItemCollection);
begin
  WriteItems(HL, GetString(TEXT_VARIABLES), V);
end;

begin
end.
