{ pas2html converts Pascal source code files to syntax-highlighted HTML files
  example:
  pas2html src\*.pas src\*.inc -o sourceashtml }
program pas2html;

uses
  filestre, objects, streams, tokenizi;

type
  TPasToHtmlConverter = object(TObject)
    OutputStream: POutputStream;
    StreamName: String;
    Tokenizer: PTokenizer;
    constructor Init(InputName, OutputName: String);
    function ConvertChar(C: Char): String;
    procedure ProcessStream;
    procedure WriteEndOfHTML;
    procedure WriteStartOfHTML;
    procedure WriteToken(Token: PToken);
  end;

function TPasToHtmlConverter.ConvertChar(C: Char): String;
begin
  case C of
    { characters already in use by HTML }
    '<': ConvertChar := '&lt;'; { less than }
    '>': ConvertChar := '&gt;'; { greater than }
    '&': ConvertChar := '&amp;'; { ampersand }
    '"': ConvertChar := '&quot;'; { quote }
    { some German special characters }
    #$e4: ConvertChar := '&auml;'; {  (small Umlaut a) }
    #$f6: ConvertChar := '&ouml;'; {  (small Umlaut o) }
    #$fc: ConvertChar := '&uuml;'; {  (small Umlaut u) }
    #$c4: ConvertChar := '&Auml;'; {  (large Umlaut A) }
    #$d6: ConvertChar := '&Ouml;'; {  (large Umlaut O) }
    #$dc: ConvertChar := '&Uuml;'; { U (large Umlaut U) }
    #$df: ConvertChar := '&szlig;'; {  (sz) }
    { more special characters to be added here }
    { otherwise, simply return the argument character }
    else ConvertChar := C;
  end;
end;

procedure TPasToHtmlConverter.ProcessPascalStream;
var
  T: PToken;
begin
  while Tokenizer^.HasData do
  begin
    T := Tokenizer^.GetToken;
    if Assigned(T) then
    begin
      WriteToken(T);
      Dispose(T, Done);
    end;
  end;
  Dispose(Tokenizer, Done);
  Tokenizer := nil;
  Dispose(OutputStream, Done);
  OutputStream := nil;
end;

procedure TPasToHtmlConverter.WriteToken(Token: PToken);
var
  C: Char;
  S: String;
begin
  if (not Assigned(T))
  then exit;
  case T^.MyType of
    TOK_COMMENT: OutputStream^.WriteStream('<i>');
    TOK_RESERVED: OutputStream^.WriteStream('<b>');
  end;
  if Assigned(T^.Data) and (T^.Data^.Content > 0) then
  for I := 0 to Data^.Content - 1 do
  begin
    C := Data^.Data^[I];
    S := ConvertChar(C);
    OutputStream^.WriteString(S);
  end;
  case T^.MyType of
    TOK_COMMENT: OutputStream^.WriteStream('</i>');
    TOK_RESERVED: OutputStream^.WriteStream('</b>');
  end;
end;

var
  OutputDirectory: String;
  StreamNames: PCollection;

procedure HandleParameters;
var
  I: Integer;
  PC: Integer;
  PS: PString;
begin
  PC := ParamCount;
  if (PC = 0) then
  begin
  	WriteLn('Usage: pas2html [-o outputdir] <FILEs>');
  	Halt(0);
  end;
  OutputDirectory := '';
  StreamNames := New(PCollection, Init(16, 16));
  I := 1;
  repeat
    S := ParamStr(I);
    if (S = '-o') then
    begin
      if (I = PC) then
      begin
      	WriteLn('Error. The -o switch must be followed by a directory name');
      	Halt(1);
      end;
      Inc(I);
      OutputDirectory := ParamStr(I);
    end
    else
    begin
      PS := NewStr(S);
      StreamNames^.Add(PS);
    end;
    Inc(I);
  until (I > PC);
end;

procedure ProcessStreams;
var
  InputStream: PFileInputStream;
  OutputStream: PFileOutputStream;
  PS: PString;
begin
  if (not Assigned(StreamNames)) or (StreamNames^.Count < 1)
  then exit;
  for I := 0 to StreamNames^.Count - 1 do
  begin
  	PS := StreamNames^.Get(I);
  	if Assigned(PS) then
  	begin
  	  InputStream := New(FileInputStream, Init(PS^));
  	  OutputStream := New(FileOutputStream, Init(OutputDirectory + PS + '.html'));
  	  Converter := New(PPascalToHtmlConverter, Init(InputStream, OutputStream));
  	  Dispose(Converter, Done);
  	end;
  end;
  Dispose(StreamNames, Done);
end;

begin
  HandleParameters;
  ProcessStreams;
end.
